import * as ts from 'typescript';
import { isLiteralExpression, isPropertyAccessExpression, isIdentifier, getCombinedModifierFlags, idText, SyntaxKind, isPropertyDeclaration, isPropertySignature, isIndexSignatureDeclaration, isMethodDeclaration, isMethodSignature, isConstructorDeclaration, isConstructSignatureDeclaration, isCallSignatureDeclaration, isTypeReferenceNode, isTypeLiteralNode, isArrayTypeNode, isTupleTypeNode, isParenthesizedTypeNode, isTypeOperatorNode, isTypePredicateNode, isUnionTypeNode, isIntersectionTypeNode, isMappedTypeNode, isConditionalTypeNode, isIndexedAccessTypeNode, isFunctionOrConstructorTypeNode, isImportTypeNode, isTypeQueryNode, isInferTypeNode, isLiteralTypeNode, isStringLiteral, isModuleBlock, isEnumDeclaration, isFunctionDeclaration, isClassDeclaration, isInterfaceDeclaration, isTypeAliasDeclaration, isModuleDeclaration, isVariableStatement, isExportDeclaration, ModifierFlags, isExportAssignment, isImportDeclaration, isNamedImports, isCallExpression, isObjectLiteralExpression, isPropertyAssignment, nodeModuleNameResolver, sys, createSourceFile, ScriptTarget } from 'typescript';

function getCodeFrame() {
    try {
        const { codeFrameColumns } = require("@babel/code-frame");
        return codeFrameColumns;
    }
    catch (_a) { }
    // istanbul ignore next
    return undefined;
}
function getLocation(node) {
    const sourceFile = node.getSourceFile();
    const start = sourceFile.getLineAndCharacterOfPosition(node.getStart());
    const end = sourceFile.getLineAndCharacterOfPosition(node.getEnd());
    return {
        start: { line: start.line + 1, column: start.character + 1 },
        end: { line: end.line + 1, column: end.character + 1 },
    };
}
function frameNodes(nodes, messages = []) {
    const codeFrame = getCodeFrame();
    const sourceFile = nodes[0].getSourceFile();
    const code = sourceFile.getFullText();
    let output = "";
    let lastLine;
    // oh jesus, why does @babel/code-frame not support this out of the box?
    for (const [i, node] of nodes.entries()) {
        const message = messages[i];
        // istanbul ignore else
        const location = getLocation(node);
        if (codeFrame) {
            const nextLocation = nodes[i + 1] && getLocation(nodes[i + 1]);
            const linesAbove = typeof lastLine === "number" ? location.start.line - lastLine - 1 : 2;
            const linesBelow = nextLocation ? nextLocation.start.line - location.end.line - 1 : 3;
            output +=
                "\n" +
                    codeFrame(code, location, {
                        highlightCode: true,
                        message,
                        linesAbove,
                        linesBelow,
                    });
            lastLine = location.end.line + linesBelow;
        }
        else {
            output += `\n${location.start.line}:${location.start.column}: \`${node.getFullText().trim()}\` <- ${message}`;
        }
    }
    return output;
}
class UnsupportedSyntaxError extends Error {
    constructor(node, message = "Syntax not yet supported") {
        super(`${message}\n${frameNodes([node])}`);
    }
}

const MARKER = "0";
let IDs = 1;
/**
 * Create a new `Program` for the given `node`:
 */
function createProgram(node) {
    return withStartEnd({
        type: "Program",
        sourceType: "module",
        body: [],
    }, node);
}
/**
 * Create a default export for `id`:
 * `export default id`
 */
function createDefaultExport(node, range) {
    return withStartEnd({
        type: "ExportDefaultDeclaration",
        declaration: createIdentifier(node),
    }, range);
}
/**
 * Create an export for `id`:
 * `export { id }`
 */
function createExport(node, range) {
    const id = createIdentifier(node);
    return withStartEnd({
        type: "ExportNamedDeclaration",
        declaration: null,
        specifiers: [
            {
                type: "ExportSpecifier",
                exported: id,
                local: id,
            },
        ],
    }, range);
}
/**
 * Creates a reference to `id`:
 * `_ = ${id}`
 */
function createReference(id) {
    return {
        type: "AssignmentPattern",
        left: {
            type: "Identifier",
            name: String(IDs++),
        },
        right: id,
    };
}
function createIdentifier(node) {
    return withStartEnd({
        type: "Identifier",
        name: node.getText(),
    }, {
        start: node.getStart(),
        end: node.getEnd(),
    });
}
/**
 * Create a new Declaration and Scope for `id`:
 * `function ${id}(_ = MARKER) {}`
 */
function createDeclaration(id, range) {
    return withStartEnd({
        type: "FunctionDeclaration",
        id: withStartEnd({
            type: "Identifier",
            name: idText(id),
        }, { start: id.getStart(), end: id.getEnd() }),
        params: [],
        body: { type: "BlockStatement", body: [] },
    }, range);
}
function convertExpression(node) {
    if (isLiteralExpression(node)) {
        return { type: "Literal", value: node.text };
    }
    if (isPropertyAccessExpression(node)) {
        return withStartEnd({
            type: "MemberExpression",
            computed: false,
            object: convertExpression(node.expression),
            property: convertExpression(node.name),
        }, {
            start: node.expression.getStart(),
            end: node.name.getEnd(),
        });
    }
    // istanbul ignore else
    if (isIdentifier(node)) {
        return createIdentifier(node);
    }
    else {
        throw new UnsupportedSyntaxError(node);
    }
}
/**
 * Mark the nested `range` to be removed, by creating dead code:
 * `_ = () => {MARKER}`
 */
function removeNested(range) {
    return createReference({
        type: "FunctionExpression",
        id: null,
        body: withStartEnd({
            type: "BlockStatement",
            body: [
                withStartEnd({
                    type: "ExpressionStatement",
                    expression: { type: "Identifier", name: MARKER },
                }, range),
            ],
        }, {
            start: getStart(range) - 1,
            end: range.end + 1,
        }),
        params: [],
    });
}
function getStart({ start, pos }) {
    return typeof start === "number" ? start : pos || 0;
}
function withStartEnd(node, range) {
    return Object.assign(node, {
        start: getStart(range),
        end: range.end,
    });
}
function matchesModifier(node, flags) {
    const nodeFlags = getCombinedModifierFlags(node);
    return (nodeFlags & flags) === flags;
}

const IGNORE_TYPENODES = new Set([
    SyntaxKind.LiteralType,
    SyntaxKind.VoidKeyword,
    SyntaxKind.UnknownKeyword,
    SyntaxKind.AnyKeyword,
    SyntaxKind.BooleanKeyword,
    SyntaxKind.NumberKeyword,
    SyntaxKind.StringKeyword,
    SyntaxKind.ObjectKeyword,
    SyntaxKind.NullKeyword,
    SyntaxKind.UndefinedKeyword,
    SyntaxKind.SymbolKeyword,
    SyntaxKind.NeverKeyword,
    SyntaxKind.ThisKeyword,
    SyntaxKind.ThisType,
]);
class DeclarationScope {
    constructor({ id, range, transformer }) {
        /**
         * As we walk the AST, we need to keep track of type variable bindings that
         * shadow the outer identifiers. To achieve this, we keep a stack of scopes,
         * represented as Sets of variable names.
         */
        this.scopes = [];
        this.transformer = transformer;
        this.declaration = createDeclaration(id, range);
    }
    pushScope() {
        this.scopes.push(new Set());
    }
    popScope(n = 1) {
        for (let i = 0; i < n; i++) {
            this.scopes.pop();
        }
    }
    pushTypeVariable(id) {
        const name = id.getText();
        this.scopes[this.scopes.length - 1].add(name);
    }
    pushRaw(expr) {
        this.declaration.params.push(expr);
    }
    pushReference(id) {
        let name;
        // We convert references from TS AST to ESTree
        // to hand them off to rollup.
        // This means we have to check the left-most identifier inside our scope
        // tree and avoid to create the reference in that case
        if (id.type === "Identifier") {
            name = id.name;
        }
        else if (id.type === "MemberExpression") {
            if (id.object.type === "Identifier") {
                name = id.object.name;
            }
        }
        if (name) {
            for (const scope of this.scopes) {
                if (scope.has(name)) {
                    return;
                }
            }
        }
        this.pushRaw(createReference(id));
    }
    pushIdentifierReference(id) {
        this.pushReference(createIdentifier(id));
    }
    removeModifier(node, kind = SyntaxKind.ExportKeyword) {
        for (const mod of node.modifiers || []) {
            if (mod.kind === kind) {
                const start = node.getStart();
                const end = mod.end + 1;
                this.pushRaw(removeNested({ start, end }));
            }
        }
    }
    convertParametersAndType(node) {
        const typeVariables = this.convertTypeParameters(node.typeParameters);
        for (const param of node.parameters) {
            this.convertTypeNode(param.type);
        }
        this.convertTypeNode(node.type);
        this.popScope(typeVariables);
    }
    convertHeritageClauses(node) {
        for (const heritage of node.heritageClauses || []) {
            for (const type of heritage.types) {
                this.pushReference(convertExpression(type.expression));
                for (const arg of type.typeArguments || []) {
                    this.convertTypeNode(arg);
                }
            }
        }
    }
    convertMembers(members) {
        for (const node of members) {
            if (isPropertyDeclaration(node) || isPropertySignature(node) || isIndexSignatureDeclaration(node)) {
                this.convertTypeNode(node.type);
                continue;
            }
            // istanbul ignore else
            if (isMethodDeclaration(node) ||
                isMethodSignature(node) ||
                isConstructorDeclaration(node) ||
                isConstructSignatureDeclaration(node) ||
                isCallSignatureDeclaration(node)) {
                this.convertParametersAndType(node);
            }
            else {
                throw new UnsupportedSyntaxError(node);
            }
        }
    }
    convertTypeParameters(params) {
        if (!params) {
            return 0;
        }
        for (const node of params) {
            this.pushScope();
            this.pushTypeVariable(node.name);
            this.convertTypeNode(node.default);
        }
        return params.length;
    }
    convertTypeNode(node) {
        if (!node) {
            return;
        }
        if (IGNORE_TYPENODES.has(node.kind)) {
            return;
        }
        if (isTypeReferenceNode(node)) {
            this.pushReference(this.convertEntityName(node.typeName));
            for (const arg of node.typeArguments || []) {
                this.convertTypeNode(arg);
            }
            return;
        }
        if (isTypeLiteralNode(node)) {
            return this.convertMembers(node.members);
        }
        if (isArrayTypeNode(node)) {
            return this.convertTypeNode(node.elementType);
        }
        if (isTupleTypeNode(node)) {
            for (const type of node.elementTypes) {
                this.convertTypeNode(type);
            }
            return;
        }
        if (isParenthesizedTypeNode(node) || isTypeOperatorNode(node) || isTypePredicateNode(node)) {
            return this.convertTypeNode(node.type);
        }
        if (isUnionTypeNode(node) || isIntersectionTypeNode(node)) {
            for (const type of node.types) {
                this.convertTypeNode(type);
            }
            return;
        }
        if (isMappedTypeNode(node)) {
            const { typeParameter, type } = node;
            this.convertTypeNode(typeParameter.constraint);
            this.pushScope();
            this.pushTypeVariable(node.typeParameter.name);
            this.convertTypeNode(type);
            this.popScope();
            return;
        }
        if (isConditionalTypeNode(node)) {
            this.convertTypeNode(node.checkType);
            this.pushScope();
            this.convertTypeNode(node.extendsType);
            this.convertTypeNode(node.trueType);
            this.convertTypeNode(node.falseType);
            this.popScope();
            return;
        }
        if (isIndexedAccessTypeNode(node)) {
            this.convertTypeNode(node.objectType);
            this.convertTypeNode(node.indexType);
            return;
        }
        if (isFunctionOrConstructorTypeNode(node)) {
            this.convertParametersAndType(node);
            return;
        }
        if (isImportTypeNode(node)) {
            this.convertImportTypeNode(node);
            return;
        }
        if (isTypeQueryNode(node)) {
            this.pushReference(this.convertEntityName(node.exprName));
            return;
        }
        // istanbul ignore else
        if (isInferTypeNode(node)) {
            this.pushTypeVariable(node.typeParameter.name);
            return;
        }
        else {
            throw new UnsupportedSyntaxError(node);
        }
    }
    // For import type nodes of the form
    // `import("./foo").Bar`
    // we create the following ESTree equivalent:
    // 1. `import * as _ from "./foo";` on the toplevel
    // 2. `_.Bar` in our declaration scope
    convertImportTypeNode(node) {
        // istanbul ignore if
        if (!isLiteralTypeNode(node.argument) || !isStringLiteral(node.argument.literal)) {
            throw new UnsupportedSyntaxError(node, "inline imports should have a literal argument");
        }
        const fileId = node.argument.literal.text;
        const start = node.getStart() + (node.isTypeOf ? "typeof ".length : 0);
        const range = { start, end: node.getEnd() };
        const importId = this.transformer.addFixupLocation(range);
        const importIdRef = withStartEnd({
            type: "Identifier",
            name: importId,
        }, range);
        this.transformer.pushStatement({
            type: "ImportDeclaration",
            specifiers: [
                {
                    type: "ImportNamespaceSpecifier",
                    local: { type: "Identifier", name: importId },
                },
            ],
            source: { type: "Literal", value: fileId },
        });
        if (node.qualifier && isIdentifier(node.qualifier)) {
            this.pushReference(withStartEnd({
                type: "MemberExpression",
                computed: false,
                object: importIdRef,
                property: createIdentifier(node.qualifier),
            }, range));
        }
        else {
            // we definitely need to do some string manipulation on the source code,
            // since rollup will not touch the `import("...")` bit at all.
            // also, for *internal* namespace references, we have the same problem
            // as with re-exporting references… -_-
            this.pushReference(importIdRef);
        }
    }
    convertEntityName(node) {
        if (isIdentifier(node)) {
            return createIdentifier(node);
        }
        return withStartEnd({
            type: "MemberExpression",
            computed: false,
            object: this.convertEntityName(node.left),
            property: createIdentifier(node.right),
        }, 
        // TODO: clean up all the `start` handling!
        { start: node.getStart(), end: node.end });
    }
    convertNamespace(node) {
        this.pushScope();
        // istanbul ignore if
        if (!node.body || !isModuleBlock(node.body)) {
            throw new UnsupportedSyntaxError(node, `namespace must have a "ModuleBlock" body.`);
        }
        const { statements } = node.body;
        // first, hoist all the declarations for correct shadowing
        for (const stmt of statements) {
            if (isEnumDeclaration(stmt) ||
                isFunctionDeclaration(stmt) ||
                isClassDeclaration(stmt) ||
                isInterfaceDeclaration(stmt) ||
                isTypeAliasDeclaration(stmt) ||
                isModuleDeclaration(stmt)) {
                // istanbul ignore else
                if (stmt.name && isIdentifier(stmt.name)) {
                    this.pushTypeVariable(stmt.name);
                }
                else {
                    throw new UnsupportedSyntaxError(stmt, `non-Identifier name not supported`);
                }
                continue;
            }
            if (isVariableStatement(stmt)) {
                for (const decl of stmt.declarationList.declarations) {
                    // istanbul ignore else
                    if (isIdentifier(decl.name)) {
                        this.pushTypeVariable(decl.name);
                    }
                    else {
                        throw new UnsupportedSyntaxError(decl, `non-Identifier name not supported`);
                    }
                }
                continue;
            }
            // istanbul ignore else
            if (isExportDeclaration(stmt)) ;
            else {
                throw new UnsupportedSyntaxError(stmt, `namespace child (hoisting) not supported yet`);
            }
        }
        // and then walk all the children like normal…
        for (const stmt of statements) {
            if (isVariableStatement(stmt)) {
                for (const decl of stmt.declarationList.declarations) {
                    if (decl.type) {
                        this.convertTypeNode(decl.type);
                    }
                }
                continue;
            }
            if (isFunctionDeclaration(stmt)) {
                this.convertParametersAndType(stmt);
                continue;
            }
            if (isInterfaceDeclaration(stmt) || isClassDeclaration(stmt)) {
                const typeVariables = this.convertTypeParameters(stmt.typeParameters);
                this.convertHeritageClauses(stmt);
                this.convertMembers(stmt.members);
                this.popScope(typeVariables);
                continue;
            }
            if (isTypeAliasDeclaration(stmt)) {
                const typeVariables = this.convertTypeParameters(stmt.typeParameters);
                this.convertTypeNode(stmt.type);
                this.popScope(typeVariables);
                continue;
            }
            if (isModuleDeclaration(stmt)) {
                this.convertNamespace(stmt);
                continue;
            }
            if (isEnumDeclaration(stmt)) {
                // noop
                continue;
            }
            // istanbul ignore else
            if (isExportDeclaration(stmt)) {
                if (stmt.exportClause) {
                    for (const decl of stmt.exportClause.elements) {
                        const id = decl.propertyName || decl.name;
                        this.pushIdentifierReference(id);
                    }
                }
            }
            else {
                throw new UnsupportedSyntaxError(stmt, `namespace child (walking) not supported yet`);
            }
        }
        this.popScope();
    }
}

class Transformer {
    constructor(sourceFile) {
        this.sourceFile = sourceFile;
        this.fixups = [];
        this.exports = new Set();
        this.ast = createProgram(sourceFile);
        for (const stmt of sourceFile.statements) {
            this.convertStatement(stmt);
        }
    }
    transform() {
        return { ast: this.ast, fixups: this.fixups };
    }
    addFixupLocation(range) {
        const original = this.sourceFile.text.slice(range.start, range.end);
        let identifier = `ɨmport${this.fixups.length}`;
        identifier += original.slice(identifier.length).replace(/[^a-zA-Z0-9_$]/g, () => "_");
        this.fixups.push({
            identifier,
            original,
            range,
        });
        return identifier;
    }
    pushStatement(node) {
        this.ast.body.push(node);
    }
    maybeMarkAsExported(node, id) {
        if (matchesModifier(node, ModifierFlags.ExportDefault)) {
            const start = node.pos;
            this.pushStatement(createDefaultExport(id, { start, end: start }));
            return true;
        }
        else if (matchesModifier(node, ModifierFlags.Export)) {
            const start = node.pos;
            const name = id.getText();
            if (this.exports.has(name)) {
                return true;
            }
            this.pushStatement(createExport(id, { start, end: start }));
            this.exports.add(name);
            return true;
        }
        return false;
    }
    createDeclaration(id, range) {
        const scope = new DeclarationScope({ id, range, transformer: this });
        this.pushStatement(scope.declaration);
        return scope;
    }
    convertStatement(node) {
        if (isEnumDeclaration(node)) {
            return this.convertEnumDeclaration(node);
        }
        if (isFunctionDeclaration(node)) {
            return this.convertFunctionDeclaration(node);
        }
        if (isInterfaceDeclaration(node) || isClassDeclaration(node)) {
            return this.convertClassOrInterfaceDeclaration(node);
        }
        if (isTypeAliasDeclaration(node)) {
            return this.convertTypeAliasDeclaration(node);
        }
        if (isVariableStatement(node)) {
            return this.convertVariableStatement(node);
        }
        if (isExportDeclaration(node) || isExportAssignment(node)) {
            return this.convertExportDeclaration(node);
        }
        if (isModuleDeclaration(node)) {
            return this.convertNamespaceDeclaration(node);
        }
        // istanbul ignore else
        if (isImportDeclaration(node)) {
            return this.convertImportDeclaration(node);
        }
        else {
            throw new UnsupportedSyntaxError(node);
        }
    }
    convertNamespaceDeclaration(node) {
        // istanbul ignore if
        if (!isIdentifier(node.name)) {
            throw new UnsupportedSyntaxError(node, `namespace name should be an "Identifier"`);
        }
        this.maybeMarkAsExported(node, node.name);
        const scope = this.createDeclaration(node.name, node);
        scope.removeModifier(node);
        scope.pushIdentifierReference(node.name);
        scope.convertNamespace(node);
    }
    convertEnumDeclaration(node) {
        this.maybeMarkAsExported(node, node.name);
        const scope = this.createDeclaration(node.name, node);
        scope.removeModifier(node);
        scope.pushIdentifierReference(node.name);
    }
    convertFunctionDeclaration(node) {
        // istanbul ignore if
        if (!node.name) {
            throw new UnsupportedSyntaxError(node, `FunctionDeclaration should have a name`);
        }
        this.maybeMarkAsExported(node, node.name);
        const scope = this.createDeclaration(node.name, node);
        scope.removeModifier(node);
        scope.pushIdentifierReference(node.name);
        scope.convertParametersAndType(node);
    }
    convertClassOrInterfaceDeclaration(node) {
        // istanbul ignore if
        if (!node.name) {
            throw new UnsupportedSyntaxError(node, `ClassDeclaration / InterfaceDeclaration should have a name`);
        }
        this.maybeMarkAsExported(node, node.name);
        const scope = this.createDeclaration(node.name, node);
        scope.removeModifier(node);
        if (isInterfaceDeclaration(node)) {
            scope.removeModifier(node, SyntaxKind.DefaultKeyword);
        }
        const typeVariables = scope.convertTypeParameters(node.typeParameters);
        scope.convertHeritageClauses(node);
        scope.convertMembers(node.members);
        scope.popScope(typeVariables);
    }
    convertTypeAliasDeclaration(node) {
        this.maybeMarkAsExported(node, node.name);
        const scope = this.createDeclaration(node.name, node);
        scope.removeModifier(node);
        const typeVariables = scope.convertTypeParameters(node.typeParameters);
        scope.convertTypeNode(node.type);
        scope.popScope(typeVariables);
    }
    convertVariableStatement(node) {
        const { declarations } = node.declarationList;
        // istanbul ignore if
        if (declarations.length !== 1) {
            throw new UnsupportedSyntaxError(node, `VariableStatement with more than one declaration not yet supported`);
        }
        for (const decl of declarations) {
            // istanbul ignore if
            if (!isIdentifier(decl.name)) {
                throw new UnsupportedSyntaxError(node, `VariableDeclaration must have a name`);
            }
            this.maybeMarkAsExported(node, decl.name);
            const scope = this.createDeclaration(decl.name, node);
            scope.removeModifier(node);
            scope.convertTypeNode(decl.type);
        }
    }
    convertExportDeclaration(node) {
        if (isExportAssignment(node)) {
            this.pushStatement(withStartEnd({
                type: "ExportDefaultDeclaration",
                declaration: convertExpression(node.expression),
            }, node));
            return;
        }
        const source = node.moduleSpecifier ? convertExpression(node.moduleSpecifier) : undefined;
        if (!node.exportClause) {
            this.pushStatement(withStartEnd({
                type: "ExportAllDeclaration",
                source,
            }, node));
        }
        else {
            const specifiers = [];
            for (const elem of node.exportClause.elements) {
                specifiers.push(this.convertExportSpecifier(elem));
            }
            this.pushStatement(withStartEnd({
                type: "ExportNamedDeclaration",
                declaration: null,
                specifiers,
                source,
            }, node));
        }
    }
    convertImportDeclaration(node) {
        const source = convertExpression(node.moduleSpecifier);
        if (!node.importClause) {
            return;
        }
        // istanbul ignore if
        if (!node.importClause.name && !node.importClause.namedBindings) {
            throw new UnsupportedSyntaxError(node, `ImportDeclaration should have imports`);
        }
        const specifiers = node.importClause.namedBindings
            ? this.convertNamedImportBindings(node.importClause.namedBindings)
            : [];
        if (node.importClause.name) {
            specifiers.push({
                type: "ImportDefaultSpecifier",
                local: createIdentifier(node.importClause.name),
            });
        }
        this.pushStatement(withStartEnd({
            type: "ImportDeclaration",
            specifiers,
            source,
        }, node));
    }
    convertNamedImportBindings(node) {
        if (isNamedImports(node)) {
            return node.elements.map(el => {
                const local = createIdentifier(el.name);
                const imported = el.propertyName ? createIdentifier(el.propertyName) : local;
                return {
                    type: "ImportSpecifier",
                    local,
                    imported,
                };
            });
        }
        return [
            {
                type: "ImportNamespaceSpecifier",
                local: createIdentifier(node.name),
            },
        ];
    }
    convertExportSpecifier(node) {
        const exported = createIdentifier(node.name);
        return {
            type: "ExportSpecifier",
            exported: exported,
            local: node.propertyName ? createIdentifier(node.propertyName) : exported,
        };
    }
}

class NamespaceFixer {
    constructor(sourceFile) {
        this.sourceFile = sourceFile;
    }
    findNamespaces() {
        const namespaces = [];
        const itemTypes = {};
        for (const node of this.sourceFile.statements) {
            if (isClassDeclaration(node)) {
                itemTypes[node.name.getText()] = "class";
            }
            else if (isFunctionDeclaration(node)) {
                itemTypes[node.name.getText()] = "function";
            }
            else if (isInterfaceDeclaration(node)) {
                itemTypes[node.name.getText()] = "interface";
            }
            else if (isModuleDeclaration(node)) {
                itemTypes[node.name.getText()] = "namespace";
            }
            else if (isEnumDeclaration(node)) {
                itemTypes[node.name.getText()] = "enum";
            }
            if (!isVariableStatement(node)) {
                continue;
            }
            const { declarations } = node.declarationList;
            if (declarations.length !== 1) {
                continue;
            }
            const decl = declarations[0];
            const name = decl.name.getText();
            if (!decl.initializer || !isCallExpression(decl.initializer)) {
                itemTypes[name] = "var";
                continue;
            }
            const obj = decl.initializer.arguments[0];
            if (!decl.initializer.expression.getFullText().includes("/*#__PURE__*/Object.freeze") ||
                !isObjectLiteralExpression(obj)) {
                continue;
            }
            const exports = [];
            for (const prop of obj.properties) {
                if (!isPropertyAssignment(prop) || !isIdentifier(prop.name) || !isIdentifier(prop.initializer)) {
                    throw new UnsupportedSyntaxError(prop, "Expected a property assignment");
                }
                exports.push({
                    exportedName: prop.name.getText(),
                    localName: prop.initializer.getText(),
                });
            }
            // sort in reverse order, since we will do string manipulation
            namespaces.unshift({
                name,
                exports,
                location: {
                    start: node.getStart(),
                    end: node.getEnd(),
                },
            });
        }
        return { namespaces, itemTypes };
    }
    fix() {
        let code = this.sourceFile.getText();
        const { namespaces, itemTypes } = this.findNamespaces();
        for (const ns of namespaces) {
            const codeAfter = code.slice(ns.location.end);
            code = code.slice(0, ns.location.start);
            for (const { exportedName, localName } of ns.exports) {
                if (exportedName === localName) {
                    const type = itemTypes[localName];
                    if (type === "interface") {
                        // an interface is just a type
                        code += `type ${ns.name}_${exportedName} = ${localName};\n`;
                    }
                    else if (type === "enum" || type === "class") {
                        // enums and classes are both types and values
                        code += `type ${ns.name}_${exportedName} = ${localName};\n`;
                        code += `declare const ${ns.name}_${exportedName}: typeof ${localName};\n`;
                    }
                    else {
                        // functions and vars are just values
                        code += `declare const ${ns.name}_${exportedName}: typeof ${localName};\n`;
                    }
                }
            }
            code += `declare namespace ${ns.name} {\n`;
            code += `  export {\n`;
            for (const { exportedName, localName } of ns.exports) {
                if (exportedName === localName) {
                    code += `    ${ns.name}_${exportedName} as ${exportedName},\n`;
                }
                else {
                    code += `    ${localName} as ${exportedName}\n`;
                }
            }
            code += `  };\n`;
            code += `}`;
            code += codeAfter;
        }
        return code;
    }
}

const plugin = () => {
    return {
        name: "dts",
        options(options) {
            return Object.assign({}, options, { treeshake: {
                    moduleSideEffects: "no-external",
                    propertyReadSideEffects: true,
                } });
        },
        outputOptions(options) {
            return Object.assign({}, options, { chunkFileNames: options.chunkFileNames || "[name]-[hash].d.ts", entryFileNames: options.entryFileNames || "[name].d.ts", format: "es", exports: "named", compact: false, freeze: true, interop: false, namespaceToStringTag: false, strict: false });
        },
        resolveId(source, importer) {
            if (!importer) {
                return;
            }
            // resolve this via typescript
            const { resolvedModule } = nodeModuleNameResolver(source, importer, {}, sys);
            if (!resolvedModule) {
                return;
            }
            // here, we define everything that comes from `node_modules` as `external`.
            // maybe its a good idea to introduce an option for this?
            if (resolvedModule.isExternalLibraryImport) {
                return { id: source, external: true };
            }
            let id = resolvedModule.resolvedFileName;
            const { extension } = resolvedModule;
            if (extension !== ".d.ts") {
                // ts resolves `.ts`/`.tsx` files before `.d.ts`
                id = id.slice(0, id.length - extension.length) + ".d.ts";
            }
            return { id };
        },
        transform(code, id) {
            if (!id.endsWith(".d.ts")) {
                this.error("`rollup-plugin-dts` can only deal with `.d.ts` files.");
                return;
            }
            const dtsSource = createSourceFile(id, code, ScriptTarget.Latest, true);
            const converter = new Transformer(dtsSource);
            const { ast, fixups } = converter.transform();
            // NOTE(swatinem):
            // hm, typescript generates `export default` without a declare,
            // but rollup moves the `export default` to a different place, which leaves
            // the function declaration without a `declare`.
            // Well luckily both words have the same length, haha :-D
            code = code.replace(/(export\s+)default(\s+(function|class))/m, "$1declare$2");
            for (const fixup of fixups) {
                code = code.slice(0, fixup.range.start) + fixup.identifier + code.slice(fixup.range.end);
            }
            return { code, ast };
        },
        renderChunk(code, chunk) {
            const source = createSourceFile(chunk.fileName, code, ScriptTarget.Latest, true);
            const fixer = new NamespaceFixer(source);
            code = fixer.fix();
            return { code, map: { mappings: "" } };
        },
    };
};

export default plugin;
